package ru.dbotthepony.kbox2d.dynamics.contact

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.collision.WorldManifold
import ru.dbotthepony.kbox2d.collision.b2TestOverlap
import ru.dbotthepony.kbox2d.dynamics.B2Body
import ru.dbotthepony.kbox2d.dynamics.B2Fixture
import java.util.*

fun interface ContactFactory {
	fun factorize(fixtureA: B2Fixture, childIndexA: Int, fixtureB: B2Fixture, childIndexB: Int): AbstractContact
}

/**
 * The class manages contact between two shapes. A contact exists for each overlapping
 * AABB in the broad-phase (except if filtered). Therefore a contact object may exist
 * that has no contact points.
 */
sealed class AbstractContact(
	/**
	 * Get fixture A in this contact.
	 */
	val fixtureA: B2Fixture,

	/**
	 * Get the child primitive index for fixture A.
	 */
	val childIndexA: Int,

	/**
	 * Get fixture B in this contact.
	 */
	val fixtureB: B2Fixture,

	/**
	 * Get the child primitive index for fixture B.
	 */
	val childIndexB: Int,
) {
	internal var flags: Int = ContactFlags.ENABLED.bitmask
	internal var isOnIsland = false
	internal var toiFlag = false

	/**
	 * Get the contact manifold. Do not modify the manifold unless you understand the
	 * internals of Box2D.
	 */
	var manifold: Manifold = Manifold()
		private set

	/**
	 * Get the next contact in the world's contact list.
	 */
	var next: AbstractContact? = null
		internal set

	/**
	 * Get the previous contact in the world's contact list.
	 */
	var prev: AbstractContact? = null
		internal set

	internal var isFlaggedForFiltering: Boolean
		get() = ContactFlags.FILTER.isit(flags)
		set(value) { flags = ContactFlags.FILTER.update(flags, value) }

	internal val nodeA: ContactEdge = ContactEdge(contact = this, other = fixtureB.body!!)
	internal val nodeB: ContactEdge = ContactEdge(contact = this, other = fixtureA.body!!)

	init {
		val bodyA = fixtureA.body as B2Body
		val bodyB = fixtureB.body as B2Body

		nodeA.next = bodyA.contactEdge
		nodeB.next = bodyB.contactEdge

		bodyA.contactEdge?.prev = nodeA
		bodyB.contactEdge?.prev = nodeB

		bodyA.contactEdge = nodeA
		bodyB.contactEdge = nodeB
	}

	/**
	 * Override the default friction mixture. You can call this in b2ContactListener::PreSolve.
	 * This value persists until set or reset.
	 * Get the friction.
	 */
	var friction: Double = b2MixFriction(fixtureA.friction, fixtureB.friction)

	/**
	 * Override the default restitution mixture. You can call this in b2ContactListener::PreSolve.
	 * The value persists until you set or reset.
	 */
	var restitution: Double = b2MixRestitution(fixtureA.restitution, fixtureB.restitution)

	/**
	 * Override the default restitution velocity threshold mixture. You can call this in b2ContactListener::PreSolve.
	 * The value persists until you set or reset.
	 * Get the restitution threshold.
	 */
	var restitutionThreshold: Double = b2MixRestitutionThreshold(fixtureA.restitutionThreshold, fixtureB.restitutionThreshold)

	/**
	 * Set the desired tangent speed for a conveyor belt behavior. In meters per second.
	 * Get the desired tangent speed. In meters per second.
	 */
	var tangentSpeed: Double = 0.0

	internal var toiCount: Int = 0
	internal var toi: Double = 0.0

	/**
	 * Evaluate this contact with your own manifold and transforms.
	 */
	abstract fun evaluate(xfA: Transform, xfB: Transform): Manifold

	/**
	 * Get the world manifold.
	 */
	val worldManifold: IWorldManifold get() {
		val bodyA = checkNotNull(fixtureA.body) { "FixtureA has no body attached" }
		val bodyB = checkNotNull(fixtureB.body) { "FixtureB has no body attached" }

		val shapeA = fixtureA.shape
		val shapeB = fixtureB.shape

		return WorldManifold(manifold, bodyA.transform, shapeA.radius, bodyB.transform, shapeB.radius)
	}

	/**
	 * Is this contact touching?
	 */
	var isTouching: Boolean
		get() = (flags and ContactFlags.TOUCHING.bitmask) == ContactFlags.TOUCHING.bitmask
		internal set(value) { flags = ContactFlags.TOUCHING.update(flags, value) }

	/**
	 * Enable/disable this contact. This can be used inside the pre-solve
	 * contact listener. The contact is only disabled for the current
	 * time step (or sub-step in continuous collisions).
	 */
	var isEnabled: Boolean
		get() = (flags and ContactFlags.ENABLED.bitmask) == ContactFlags.ENABLED.bitmask
		set(value) { flags = ContactFlags.ENABLED.update(flags, value) }

	fun flagForFiltering() {
		flags = flags or ContactFlags.FILTER.bitmask
	}

	/**
	 * Awakens fixtures this contact belongs to if neither of them are sensors
	 */
	fun destroy() {
		if (manifold.points.isNotEmpty() && !fixtureA.isSensor && !fixtureB.isSensor) {
			fixtureA.body!!.isAwake = true
			fixtureB.body!!.isAwake = true
		}
	}

	/**
	 * Update the contact manifold and touching status.
	 * Note: do not assume the fixture AABBs are overlapping or are valid.
	 */
	internal fun update(contactListener: IContactListener?) {
		// Re-enable this contact.
		flags = ContactFlags.ENABLED.or(flags)

		val oldManifold = manifold

		val touching: Boolean
		val wasTouching = isTouching

		val sensor = fixtureA.isSensor || fixtureB.isSensor

		val bodyA = fixtureA.body as B2Body
		val bodyB = fixtureB.body as B2Body

		val xfA = bodyA.transform
		val xfB = bodyB.transform

		// Is this contact a sensor?
		if (sensor) {
			val shapeA = fixtureA.shape
			val shapeB = fixtureB.shape

			touching = b2TestOverlap(shapeA, childIndexA, shapeB, childIndexB, xfA, xfB)
			// Sensors don't generate manifolds.
			manifold = Manifold.EMPTY
		} else {
			manifold = evaluate(xfA, xfB)
			touching = manifold.points.isNotEmpty()

			// Match old contact ids to new contact ids and copy the
			// stored impulses to warm start the solver.
			for (mp2 in manifold.points) {
				mp2.normalImpulse = 0.0
				mp2.tangentImpulse = 0.0
				val id2 = mp2.id

				for (mp1 in oldManifold.points) {
					if (mp1.id.key == id2.key) {
						mp2.normalImpulse = mp1.normalImpulse
						mp2.tangentImpulse = mp1.tangentImpulse
						break
					}
				}
			}

			if (touching != wasTouching) {
				bodyA.isAwake = true
				bodyB.isAwake = true
			}
		}

		this.isTouching = touching

		if (!wasTouching && touching) {
			contactListener?.beginContact(this)
		} else if (wasTouching && !touching) {
			contactListener?.endContact(this)
		}

		if (!sensor && touching) {
			contactListener?.preSolve(this, oldManifold)
		}
	}

	/**
	 * Reset the friction mixture to the default value.
	 */
	fun resetFriction() {
		friction = b2MixFriction(fixtureA.friction, fixtureB.friction)
	}

	/**
	 * Reset the restitution to the default value.
 	 */
	fun resetRestitution() {
		restitution = b2MixRestitution(fixtureA.restitution, fixtureB.restitution)
	}

	/**
	 * Reset the restitution threshold to the default value.
	 */
	fun resetRestitutionThreshold() {
		restitutionThreshold = b2MixRestitutionThreshold(fixtureA.restitutionThreshold, fixtureB.restitutionThreshold)
	}

	companion object {
		private val registry =
			EnumMap<IShape.Type, EnumMap<IShape.Type, ContactFactory>>(IShape.Type::class.java)

		internal fun register(type1: IShape.Type, type2: IShape.Type, factory: ContactFactory) {
			registry.computeIfAbsent(type1) { EnumMap(IShape.Type::class.java) }[type2] = factory
		}

		internal fun create(
			fixtureA: B2Fixture,
			indexA: Int,
			fixtureB: B2Fixture,
			indexB: Int,
		): AbstractContact {
			val type1 = fixtureA.type
			val type2 = fixtureB.type

			return registry[type1]?.get(type2)?.factorize(fixtureA, indexA, fixtureB, indexB)
				?: registry[type2]?.get(type1)?.factorize(fixtureB, indexB, fixtureA, indexA)
				?: throw IllegalArgumentException("No collision handler for between $type1 and $type2")
		}

		init {
			register(IShape.Type.POLYGON, IShape.Type.POLYGON) { fixtureA: B2Fixture, _: Int, fixtureB: B2Fixture, _: Int ->
				return@register PolygonContact(fixtureA, fixtureB)
			}

			register(IShape.Type.POLYGON, IShape.Type.CIRCLE) { fixtureA: B2Fixture, _: Int, fixtureB: B2Fixture, _: Int ->
				return@register PolygonCircleContact(fixtureA, fixtureB)
			}

			register(IShape.Type.CIRCLE, IShape.Type.CIRCLE) { fixtureA: B2Fixture, _: Int, fixtureB: B2Fixture, _: Int ->
				return@register CircleContact(fixtureA, fixtureB)
			}

			register(IShape.Type.EDGE, IShape.Type.CIRCLE) { fixtureA: B2Fixture, _: Int, fixtureB: B2Fixture, _: Int ->
				return@register EdgeCircleContact(fixtureA, fixtureB)
			}

			register(IShape.Type.EDGE, IShape.Type.POLYGON) { fixtureA: B2Fixture, _: Int, fixtureB: B2Fixture, _: Int ->
				return@register EdgePolygonContact(fixtureA, fixtureB)
			}

			register(IShape.Type.CHAIN, IShape.Type.POLYGON) { fixtureA: B2Fixture, indexA: Int, fixtureB: B2Fixture, indexB: Int ->
				return@register ChainPolygonContact(fixtureA, indexA, fixtureB, indexB)
			}

			register(IShape.Type.CHAIN, IShape.Type.CIRCLE) { fixtureA: B2Fixture, indexA: Int, fixtureB: B2Fixture, indexB: Int ->
				return@register ChainCircleContact(fixtureA, indexA, fixtureB, indexB)
			}
		}
	}
}
