
@file:Suppress("unused")

package ru.dbotthepony.kvector.matrix.nfloat

import ru.dbotthepony.kvector.api.IMutableSquareMatrix
import ru.dbotthepony.kvector.api.concrete.IMatrix4f
import ru.dbotthepony.kvector.matrix.ndouble.Matrix4d
import ru.dbotthepony.kvector.matrix.ndouble.MutableMatrix4d
import ru.dbotthepony.kvector.narray.Float2Dimensional
import ru.dbotthepony.kvector.vector.nfloat.*
import kotlin.math.tan

/**
 * Represents immutable concrete matrix with 4x4 dimensions, storing values as [Float]s.
 *
 * See [Matrix4d] for documentation, with reflection of this class.
 *
 * Vectorized access use [Vector4f], generic interface is represented by [IMatrix4f].
 */
class Matrix4f : AbstractMatrixVf<Matrix4f>, IMatrix4f<Matrix4f> {
	constructor(
		c00: Float = 1f, c01: Float = 0f, c02: Float = 0f, c03: Float = 0f,
		c10: Float = 0f, c11: Float = 1f, c12: Float = 0f, c13: Float = 0f,
		c20: Float = 0f, c21: Float = 0f, c22: Float = 1f, c23: Float = 0f,
		c30: Float = 0f, c31: Float = 0f, c32: Float = 0f, c33: Float = 1f,
	) : super(4, 4, Float2Dimensional(4, 4).also {
		it[0, 0] = c00
		it[1, 0] = c10
		it[2, 0] = c20
		it[3, 0] = c30

		it[0, 1] = c01
		it[1, 1] = c11
		it[2, 1] = c21
		it[3, 1] = c31

		it[0, 2] = c02
		it[1, 2] = c12
		it[2, 2] = c22
		it[3, 2] = c32

		it[0, 3] = c03
		it[1, 3] = c13
		it[2, 3] = c23
		it[3, 3] = c33
	})

	private constructor(memory: Float2Dimensional) : super(4, 4, memory)

	override val flexible: Boolean = false

	override fun factorize(memory: Float2Dimensional): Matrix4f {
		requireSizeEquals(memory)
		return Matrix4f(memory)
	}

	override val c00: Float get() = memory[0, 0]
	override val c10: Float get() = memory[1, 0]
	override val c20: Float get() = memory[2, 0]
	override val c30: Float get() = memory[3, 0]
	override val c01: Float get() = memory[0, 1]
	override val c11: Float get() = memory[1, 1]
	override val c21: Float get() = memory[2, 1]
	override val c31: Float get() = memory[3, 1]
	override val c02: Float get() = memory[0, 2]
	override val c12: Float get() = memory[1, 2]
	override val c22: Float get() = memory[2, 2]
	override val c32: Float get() = memory[3, 2]
	override val c03: Float get() = memory[0, 3]
	override val c13: Float get() = memory[1, 3]
	override val c23: Float get() = memory[2, 3]
	override val c33: Float get() = memory[3, 3]

	override val r00 by this::c00
	override val r01 by this::c10
	override val r02 by this::c20
	override val r03 by this::c30
	override val r10 by this::c01
	override val r11 by this::c11
	override val r12 by this::c21
	override val r13 by this::c31
	override val r20 by this::c02
	override val r21 by this::c12
	override val r22 by this::c22
	override val r23 by this::c32
	override val r30 by this::c03
	override val r31 by this::c13
	override val r32 by this::c23
	override val r33 by this::c33

	override val a11 by this::c00
	override val a21 by this::c01
	override val a31 by this::c02
	override val a41 by this::c03
	override val a12 by this::c10
	override val a22 by this::c11
	override val a32 by this::c12
	override val a42 by this::c13
	override val a13 by this::c20
	override val a23 by this::c21
	override val a33 by this::c22
	override val a43 by this::c23
	override val a14 by this::c30
	override val a24 by this::c31
	override val a34 by this::c32
	override val a44 by this::c33

	override val b11 by this::c00
	override val b21 by this::c10
	override val b31 by this::c20
	override val b41 by this::c30
	override val b12 by this::c01
	override val b22 by this::c11
	override val b32 by this::c21
	override val b42 by this::c31
	override val b13 by this::c02
	override val b23 by this::c12
	override val b33 by this::c22
	override val b43 by this::c32
	override val b14 by this::c03
	override val b24 by this::c13
	override val b34 by this::c23
	override val b44 by this::c33

	override val c0: Vector4f get() = Vector4f(c00, c01, c02, c03)
	override val c1: Vector4f get() = Vector4f(c10, c11, c12, c13)
	override val c2: Vector4f get() = Vector4f(c20, c21, c22, c23)
	override val c3: Vector4f get() = Vector4f(c30, c31, c32, c33)

	override val r0: Vector4f get() = Vector4f(r00, r01, r02, r03)
	override val r1: Vector4f get() = Vector4f(r10, r11, r12, r13)
	override val r2: Vector4f get() = Vector4f(r20, r21, r22, r23)
	override val r3: Vector4f get() = Vector4f(r30, r31, r32, r33)

	override fun translate(vector: Vector3f): Matrix4f {
		return rm(
			r00, r01, r02, r03 + vector.x,
			r10, r11, r12, r13 + vector.y,
			r20, r21, r22, r23 + vector.z,
			r30, r31, r32, r33,
		)
	}

	override fun translateWithMultiplication(vector: Vector3f): Matrix4f {
		return rm(
			r00, r01, r02, r03 + vector.x * r00 + vector.y * r01 + vector.z * r02,
			r10, r11, r12, r13 + vector.x * r10 + vector.y * r11 + vector.z * r12,
			r20, r21, r22, r23 + vector.x * r20 + vector.y * r21 + vector.z * r22,
			r30, r31, r32, r33,
		)
	}

	override fun scale(vector: Vector4f): Matrix4f {
		return rm(
			r00 * vector.x, r01, r02, r03,
			r10, r11 * vector.y, r12, r13,
			r20, r21, r22 * vector.z, r23,
			r30, r31, r32, r33 * vector.w,
		)
	}

	companion object {
		/**
		 * Initializes new matrix, with arguments taken in *Row-Major* order, that is, fist row of values
		 * is fist row in created matrix.
		 */
		fun rm(
			r00: Float = 1f, r01: Float = 0f, r02: Float = 0f, r03: Float = 0f,
			r10: Float = 0f, r11: Float = 1f, r12: Float = 0f, r13: Float = 0f,
			r20: Float = 0f, r21: Float = 0f, r22: Float = 1f, r23: Float = 0f,
			r30: Float = 0f, r31: Float = 0f, r32: Float = 0f, r33: Float = 1f,
		): Matrix4f {
			return Matrix4f(
				c00 = r00, c10 = r01, c20 = r02, c30 = r03,
				c01 = r10, c11 = r11, c21 = r12, c31 = r13,
				c02 = r20, c12 = r21, c22 = r22, c32 = r23,
				c03 = r30, c13 = r31, c23 = r32, c33 = r33,
			)
		}

		/**
		 * Initializes new zero matrix.
		 *
		 * Since [Matrix4f] is immutable, it is encouraged you use [ZERO] property instead.
		 */
		fun zero(): Matrix4f {
			return Matrix4f(
				c00 = 0f, c10 = 0f, c20 = 0f, c30 = 0f,
				c01 = 0f, c11 = 0f, c21 = 0f, c31 = 0f,
				c02 = 0f, c12 = 0f, c22 = 0f, c32 = 0f,
				c03 = 0f, c13 = 0f, c23 = 0f, c33 = 0f,
			)
		}

		/**
		 * All values of this matrix are zero
		 */
		@JvmField val ZERO = zero()

		/**
		 * Main diagonal of this matrix is 1s
		 */
		@JvmField val IDENTITY = Matrix4f()

		/**
		 * Returns new ortho projection matrix, with Y coordinate inverted (useful for OpenGL GUI drawing).
		 *
		 * Inversion of Y means that X 0 Y 0 will be top left position on screen (in OpenGL), not bottom left.
		 */
		fun ortho(left: Float, right: Float, bottom: Float, top: Float, zNear: Float, zFar: Float): Matrix4f {
			return rm(
				r00 = 2f / (right - left),
				r11 = -2f / (top - bottom),
				r22 = 2f / (zFar - zNear),
				r03 = -(right + left) / (right - left),
				r13 = -(top + bottom) / (top - bottom) + 2f,
				r23 = -(zFar + zNear) / (zFar - zNear)
			)
		}

		/**
		 * Returns new ortho projection matrix.
		 */
		fun orthoDirect(left: Float, right: Float, bottom: Float, top: Float, zNear: Float, zFar: Float): Matrix4f {
			return rm(
				r00 = 2f / (right - left),
				r11 = 2f / (top - bottom),
				r22 = 2f / (zFar - zNear),
				r03 = -(right + left) / (right - left),
				r13 = -(top + bottom) / (top - bottom),
				r23 = -(zFar + zNear) / (zFar - zNear)
			)
		}

		/**
		 * Returns new perspective matrix
		 */
		fun perspective(fov: Float, zFar: Float, zNear: Float): Matrix4f {
			val scale = (1.0 / (tan(Math.toRadians(fov.toDouble()) / 2.0))).toFloat()
			val r = zFar - zNear

			return rm(
				r00 = scale,
				r11 = scale,
				r22 = -zFar / r,
				r23 = -1f,
				r32 = -zFar * zNear / r,
			)
		}
	}
}

/**
 * Represents mutable concrete matrix with 4x4 dimensions, storing values as [Float]s.
 *
 * See [MutableMatrix4d] for documentation, with reflection of this class.
 *
 * Vectorized access use [MutableVector4f], generic interface is represented by [IMatrix4f].
 */
class MutableMatrix4f : AbstractMutableMatrixVf<MutableMatrix4f>, IMatrix4f<MutableMatrix4f>, IMutableSquareMatrix<MutableMatrix4f, Vector3f, Vector4f> {
	constructor(
		c00: Float = 1f, c01: Float = 0f, c02: Float = 0f, c03: Float = 0f,
		c10: Float = 0f, c11: Float = 1f, c12: Float = 0f, c13: Float = 0f,
		c20: Float = 0f, c21: Float = 0f, c22: Float = 1f, c23: Float = 0f,
		c30: Float = 0f, c31: Float = 0f, c32: Float = 0f, c33: Float = 1f,
	) : super(4, 4, Float2Dimensional(4, 4).also {
		it[0, 0] = c00
		it[1, 0] = c10
		it[2, 0] = c20
		it[3, 0] = c30

		it[0, 1] = c01
		it[1, 1] = c11
		it[2, 1] = c21
		it[3, 1] = c31

		it[0, 2] = c02
		it[1, 2] = c12
		it[2, 2] = c22
		it[3, 2] = c32

		it[0, 3] = c03
		it[1, 3] = c13
		it[2, 3] = c23
		it[3, 3] = c33
	})

	private constructor(memory: Float2Dimensional) : super(4, 4, memory)

	override val flexible: Boolean = false

	override fun factorize(memory: Float2Dimensional): MutableMatrix4f {
		requireSizeEquals(memory)
		return MutableMatrix4f(memory)
	}

	override fun translate(vector: Vector3f): MutableMatrix4f {
		return rm(
			r00, r01, r02, r03 + vector.x,
			r10, r11, r12, r13 + vector.y,
			r20, r21, r22, r23 + vector.z,
			r30, r31, r32, r33,
		)
	}

	override fun translateMut(vector: Vector3f): MutableMatrix4f {
		r03 += vector.x
		r13 += vector.y
		r23 += vector.z
		return this
	}

	override fun translateWithMultiplication(vector: Vector3f): MutableMatrix4f {
		return rm(
			r00, r01, r02, r03 + vector.x * r00 + vector.y * r01 + vector.z * r02,
			r10, r11, r12, r13 + vector.x * r10 + vector.y * r11 + vector.z * r12,
			r20, r21, r22, r23 + vector.x * r20 + vector.y * r21 + vector.z * r22,
			r30, r31, r32, r33,
		)
	}

	override fun scale(vector: Vector4f): MutableMatrix4f {
		return rm(
			r00 * vector.x, r01, r02, r03,
			r10, r11 * vector.y, r12, r13,
			r20, r21, r22 * vector.z, r23,
			r30, r31, r32, r33 * vector.w,
		)
	}

	override fun translateWithMultiplicationMut(vector: Vector3f): MutableMatrix4f {
		r03 += vector.x * r00 + vector.y * r01 + vector.z * r02
		r13 += vector.x * r10 + vector.y * r11 + vector.z * r12
		r23 += vector.x * r20 + vector.y * r21 + vector.z * r22
		return this
	}

	override fun scaleMut(vector: Vector4f): MutableMatrix4f {
		c00 *= vector.x
		c11 *= vector.y
		c22 *= vector.z
		c33 *= vector.w
		return this
	}

	override var translation: Vector3f
		get() = super.translation
		set(value) {
			r03 = value.x
			r13 = value.y
			r23 = value.z
		}

	override var c00: Float
		get() = memory[0, 0]
		set(value) { memory[0, 0] = value }
	override var c10: Float
		get() = memory[1, 0]
		set(value) { memory[1, 0] = value }
	override var c20: Float
		get() = memory[2, 0]
		set(value) { memory[2, 0] = value }
	override var c30: Float
		get() = memory[3, 0]
		set(value) { memory[3, 0] = value }
	override var c01: Float
		get() = memory[0, 1]
		set(value) { memory[0, 1] = value }
	override var c11: Float
		get() = memory[1, 1]
		set(value) { memory[1, 1] = value }
	override var c21: Float
		get() = memory[2, 1]
		set(value) { memory[2, 1] = value }
	override var c31: Float
		get() = memory[3, 1]
		set(value) { memory[3, 1] = value }
	override var c02: Float
		get() = memory[0, 2]
		set(value) { memory[0, 2] = value }
	override var c12: Float
		get() = memory[1, 2]
		set(value) { memory[1, 2] = value }
	override var c22: Float
		get() = memory[2, 2]
		set(value) { memory[2, 2] = value }
	override var c32: Float
		get() = memory[3, 2]
		set(value) { memory[3, 2] = value }
	override var c03: Float
		get() = memory[0, 3]
		set(value) { memory[0, 3] = value }
	override var c13: Float
		get() = memory[1, 3]
		set(value) { memory[1, 3] = value }
	override var c23: Float
		get() = memory[2, 3]
		set(value) { memory[2, 3] = value }
	override var c33: Float
		get() = memory[3, 3]
		set(value) { memory[3, 3] = value }

	override var r00 by this::c00
	override var r01 by this::c10
	override var r02 by this::c20
	override var r03 by this::c30
	override var r10 by this::c01
	override var r11 by this::c11
	override var r12 by this::c21
	override var r13 by this::c31
	override var r20 by this::c02
	override var r21 by this::c12
	override var r22 by this::c22
	override var r23 by this::c32
	override var r30 by this::c03
	override var r31 by this::c13
	override var r32 by this::c23
	override var r33 by this::c33

	override var a11 by this::c00
	override var a21 by this::c01
	override var a31 by this::c02
	override var a41 by this::c03
	override var a12 by this::c10
	override var a22 by this::c11
	override var a32 by this::c12
	override var a42 by this::c13
	override var a13 by this::c20
	override var a23 by this::c21
	override var a33 by this::c22
	override var a43 by this::c23
	override var a14 by this::c30
	override var a24 by this::c31
	override var a34 by this::c32
	override var a44 by this::c33

	override var b11 by this::c00
	override var b21 by this::c01
	override var b31 by this::c02
	override var b41 by this::c03
	override var b12 by this::c10
	override var b22 by this::c11
	override var b32 by this::c12
	override var b42 by this::c13
	override var b13 by this::c20
	override var b23 by this::c21
	override var b33 by this::c22
	override var b43 by this::c23
	override var b14 by this::c30
	override var b24 by this::c31
	override var b34 by this::c32
	override var b44 by this::c33

	override var c0: Vector4f
		get() = object : MutableVector4f(c00, c01, c02, c03) {
			override var x by this@MutableMatrix4f::c00
			override var y by this@MutableMatrix4f::c01
			override var z by this@MutableMatrix4f::c02
			override var w by this@MutableMatrix4f::c03
		}
		set(value) {
			c00 = value.x
			c01 = value.y
			c02 = value.z
			c03 = value.w
		}

	override var c1: Vector4f
		get() = object : MutableVector4f(c10, c11, c12, c13) {
			override var x by this@MutableMatrix4f::c10
			override var y by this@MutableMatrix4f::c11
			override var z by this@MutableMatrix4f::c12
			override var w by this@MutableMatrix4f::c13
		}
		set(value) {
			c10 = value.x
			c11 = value.y
			c12 = value.z
			c13 = value.w
		}

	override var c2: Vector4f
		get() = object : MutableVector4f(c20, c21, c22, c23) {
			override var x by this@MutableMatrix4f::c20
			override var y by this@MutableMatrix4f::c21
			override var z by this@MutableMatrix4f::c22
			override var w by this@MutableMatrix4f::c23
		}
		set(value) {
			c20 = value.x
			c21 = value.y
			c22 = value.z
			c23 = value.w
		}

	override var c3: Vector4f
		get() = object : MutableVector4f(c30, c31, c32, c33) {
			override var x by this@MutableMatrix4f::c30
			override var y by this@MutableMatrix4f::c31
			override var z by this@MutableMatrix4f::c32
			override var w by this@MutableMatrix4f::c33
		}
		set(value) {
			c30 = value.x
			c31 = value.y
			c32 = value.z
			c33 = value.w
		}

	override var r0: Vector4f
		get() = object : MutableVector4f(r00, r01, r02, r03) {
			override var x by this@MutableMatrix4f::r00
			override var y by this@MutableMatrix4f::r01
			override var z by this@MutableMatrix4f::r02
			override var w by this@MutableMatrix4f::r03
		}
		set(value) {
			r00 = value.x
			r01 = value.y
			r02 = value.z
			r03 = value.w
		}

	override var r1: Vector4f
		get() = object : MutableVector4f(r10, r11, r12, r13) {
			override var x by this@MutableMatrix4f::r10
			override var y by this@MutableMatrix4f::r11
			override var z by this@MutableMatrix4f::r12
			override var w by this@MutableMatrix4f::r13
		}
		set(value) {
			r10 = value.x
			r11 = value.y
			r12 = value.z
			r13 = value.w
		}

	override var r2: Vector4f
		get() = object : MutableVector4f(r20, r21, r22, r23) {
			override var x by this@MutableMatrix4f::r20
			override var y by this@MutableMatrix4f::r21
			override var z by this@MutableMatrix4f::r22
			override var w by this@MutableMatrix4f::r23
		}
		set(value) {
			r20 = value.x
			r21 = value.y
			r22 = value.z
			r23 = value.w
		}

	override var r3: Vector4f
		get() = object : MutableVector4f(r30, r31, r32, r33) {
			override var x by this@MutableMatrix4f::r30
			override var y by this@MutableMatrix4f::r31
			override var z by this@MutableMatrix4f::r32
			override var w by this@MutableMatrix4f::r33
		}
		set(value) {
			r30 = value.x
			r31 = value.y
			r32 = value.z
			r33 = value.w
		}

	companion object {
		/**
		 * Initializes new matrix, with arguments taken in *Row-Major* order, that is, fist row of values
		 * is fist row in created matrix.
		 */
		fun rm(
			r00: Float = 1f, r01: Float = 0f, r02: Float = 0f, r03: Float = 0f,
			r10: Float = 0f, r11: Float = 1f, r12: Float = 0f, r13: Float = 0f,
			r20: Float = 0f, r21: Float = 0f, r22: Float = 1f, r23: Float = 0f,
			r30: Float = 0f, r31: Float = 0f, r32: Float = 0f, r33: Float = 1f,
		): MutableMatrix4f {
			return MutableMatrix4f(
				c00 = r00, c10 = r01, c20 = r02, c30 = r03,
				c01 = r10, c11 = r11, c21 = r12, c31 = r13,
				c02 = r20, c12 = r21, c22 = r22, c32 = r23,
				c03 = r30, c13 = r31, c23 = r32, c33 = r33,
			)
		}

		/**
		 * Initializes new mutable zero matrix.
		 */
		fun zero(): MutableMatrix4f {
			return MutableMatrix4f(
				c00 = 0f, c10 = 0f, c20 = 0f, c30 = 0f,
				c01 = 0f, c11 = 0f, c21 = 0f, c31 = 0f,
				c02 = 0f, c12 = 0f, c22 = 0f, c32 = 0f,
				c03 = 0f, c13 = 0f, c23 = 0f, c33 = 0f,
			)
		}

		/**
		 * Returns new ortho projection matrix, with Y coordinate inverted (useful for OpenGL GUI drawing).
		 *
		 * Inversion of Y means that X 0 Y 0 will be top left position on screen (in OpenGL), not bottom left.
		 */
		fun ortho(left: Float, right: Float, bottom: Float, top: Float, zNear: Float, zFar: Float): MutableMatrix4f {
			return rm(
				r00 = 2f / (right - left),
				r11 = -2f / (top - bottom),
				r22 = 2f / (zFar - zNear),
				r03 = -(right + left) / (right - left),
				r13 = -(top + bottom) / (top - bottom) + 2f,
				r23 = -(zFar + zNear) / (zFar - zNear)
			)
		}

		/**
		 * Returns new ortho projection matrix.
		 */
		fun orthoDirect(left: Float, right: Float, bottom: Float, top: Float, zNear: Float, zFar: Float): MutableMatrix4f {
			return rm(
				r00 = 2f / (right - left),
				r11 = 2f / (top - bottom),
				r22 = 2f / (zFar - zNear),
				r03 = -(right + left) / (right - left),
				r13 = -(top + bottom) / (top - bottom),
				r23 = -(zFar + zNear) / (zFar - zNear)
			)
		}

		/**
		 * Returns new perspective matrix
		 */
		fun perspective(fov: Float, zFar: Float, zNear: Float): MutableMatrix4f {
			val scale = (1.0 / (tan(Math.toRadians(fov.toDouble()) / 2.0))).toFloat()
			val r = zFar - zNear

			return rm(
				r00 = scale,
				r11 = scale,
				r22 = -zFar / r,
				r23 = -1f,
				r32 = -zFar * zNear / r,
			)
		}
	}
}